#include "simple3Layers.h"

/* -------------------------------------------------------------------------------------- */

void  Simple3Layers::read_from_file ( void ) { 
  ifstream I ("../CONTROLLERS/simple3Layers");
  if(!I) {
    cerr << "File with network structure not found" <<endl;
    exit(0);  
  }
  
  num_input               = getInt      ('=', I);
  num_hidden              = getInt      ('=', I);
  num_output              = getInt      ('=', I);
  low_bound_Wts           = getDouble   ('=', I);
  upper_bound_Wts         = getDouble   ('=', I);
  
  int check = getInt('=',I);
  I.close();
  if( check != 999 ){
    cerr << " In ../CONTROLLERS/simple3layers specifications' file --- End check was not 999" << endl;
    exit(0);
  }
}

/* -------------------------------------------------------------------------------------- */

void Simple3Layers::compute_genotype_length ( void ){
  genotype_length = 0;
  //weights input-hidden
  for(int i = 0; i < num_input; i++) 
    for(int h = 0; h < num_hidden; h++) 
      genotype_length++;
  
  //weights hidden-output
  for(int h = 0; h < num_hidden; h++) 
    for(int o = 0; o < num_output; o++) 
      genotype_length++;
  
  //weigth bias-hidden 
  for(int h = 0; h < num_hidden; h++) genotype_length++;
  
  //weigths bias-output
  for(int out = 0; out < num_output; out++) genotype_length++;
}

/* -------------------------------------------------------------------------------------- */

Simple3Layers::Simple3Layers( ) : Controller() {  
  read_from_file ( );
  compute_genotype_length ( );
  
  inputLayer.resize   ( num_input );
  hiddenLayer.resize  ( num_hidden );
  outputLayer.resize  ( num_output );
  biasLayer.resize    ( 1 );
  
  for( int i=0; i<num_input; i++) {
    inputLayer[i].weightsOut    = new double[num_hidden];
  }
  for( int i=0; i<num_hidden; i++) {
    hiddenLayer[i].weightsOut   = new double[num_output];
  }
  for( int i=0; i<num_output; i++){
    outputLayer[i].weightsOut  = NULL;
  }
  biasLayer[0].weightsOut   = new double[num_hidden+num_output];
}

/* -------------------------------------------------------------------------------------- */

Simple3Layers::~Simple3Layers(){
  for(  int i=0; i<inputLayer.size(); i++) {
    delete[] inputLayer[i].weightsOut;
  }
  for(  int i=0; i<hiddenLayer.size(); i++){
    delete[] hiddenLayer[i].weightsOut;
  }
  for(  int i=0; i<outputLayer.size(); i++) {
    delete[] outputLayer[i].weightsOut;
  }
  delete[] biasLayer[0].weightsOut;
}

/* -------------------------------------------------------------------------------------- */

void Simple3Layers::init( const vector <chromosome_type> &genes ){
  int    counter  = 0;
  // /* ------------------ INPUT_LAYER -------------------- */
  for( int i=0; i<num_input; i++){
    inputLayer[i].state        = 0.0;
    for(int j=0; j< num_hidden; j++){
      inputLayer[i].weightsOut[j] = get_value(genes, counter)*(upper_bound_Wts - low_bound_Wts) + low_bound_Wts;
      counter++;
    }
  }
  
  // /* ------------------ CONTROLLER::HIDDEN_LAYER -------------------- */
  for( int i=0; i<num_hidden; i++){
    hiddenLayer[i].state        = 0.0;
    for(int j=0; j<num_output; j++){
      hiddenLayer[i].weightsOut[j] = get_value(genes, counter)*(upper_bound_Wts - low_bound_Wts) + low_bound_Wts;
      counter++;
    }
  }
  
  // /* ------------------ CONTROLLER::OUTPUT_LAYER -------------------- */
  for(int i=0; i<num_output; i++)
    outputLayer[i].state         = 0.0;
  
  // /* ------------------ CONTROLLER::BIAS_LAYER -------------------- */
  biasLayer[0].state         = 1.0;
  for(int i=0; i<(num_hidden+num_output); i++){
    biasLayer[0].weightsOut[i] = get_value(genes, counter)*(upper_bound_Wts - low_bound_Wts) + low_bound_Wts;
    counter++;
  }
  
  // /* --------------------------------------------------------------- */
  if( counter != genotype_length ){
    cerr << "'In simple3Layers.cpp init():: the number of genes is wrong"
   	 << " " << genotype_length << " " << counter << endl;
    exit(0);
  }
}

/* -------------------------------------------------------------------------------------- */

void Simple3Layers::step ( const vector <double> &inputs, vector <double> &outputs ){
  //int w=0;
 /* for( int i = 0; i < get_num_input(); i++){
           
           printf("ir%d reading= %f\n",i,inputs[i]);
      } */

 
  for( int h=0; h<num_hidden; h++){
    hiddenLayer[h].state = biasLayer[0].weightsOut[h];
    //printf("weight[%d]= %lf    ",w,biasLayer[0].weightsOut[h]);
    //w++;
    for( int i=0; i<num_input; i++){
      hiddenLayer[h].state += inputs[i] * inputLayer[i].weightsOut[h];
      //printf("weight[%d]= %lf    ",w,inputLayer[i].weightsOut[h]);
      //w++;
    }
    hiddenLayer[h].state = f_sigmoid( hiddenLayer[h].state );
  }
  
  for(int out=0; out<num_output; out++){
    outputLayer[out].state = biasLayer[0].weightsOut[num_hidden+out];
    //printf("weight[%d]= %lf    ",w,biasLayer[0].weightsOut[num_hidden+out]);
    //w++;
    for( int h=0; h<num_hidden; h++){
      outputLayer[out].state += hiddenLayer[h].state * hiddenLayer[h].weightsOut[out];
     // printf("weight[%d]= %lf    ",w,hiddenLayer[h].weightsOut[out]);
     // w++;
    }
    outputs[out] = f_sigmoid( outputLayer[out].state );
  }
// printf("\n net out0=%f  out1=%f out2=%f  out3=%f\n",outputs[0],outputs[1],outputs[2],outputs[3]);
}

/* -------------------------------------------------------------------------------------- */

void Simple3Layers::reset ( void ){
  for(int i=0; i<num_input; i++){
    inputLayer[i].state = 0.0;
  }
  for(int i=0; i<num_hidden; i++){
    hiddenLayer[i].state = 0.0;
  }
  for(int i=0; i<num_output; i++){
    outputLayer[i].state = 0.0;
  }
}

/* -------------------------------------------------------------------------------------- */
