#include "experiment.h"
#include "../EVOL_ALGORITHM/ga.h"
//#include <stdio.h>
//#include <conio.h>
/* ---------------------------------------------------------------------------------------- */

EXP_Class::EXP_Class(const char *run_name, bool _evolution, bool _viewing, bool _re_evaluation, unsigned long _seed, int _from_gen, int _to_gen, int _num_cores ){

    evolution     = _evolution;
    viewing       = _viewing;
    re_evaluation = _re_evaluation;
    from_gen = _from_gen;
    to_gen   = _to_gen;
    num_core = _num_cores;
    param     = new Parameters ( run_name );
    if( evolution ) param->set_random_root_seed( _seed );
    param->init_random_generator( );//GSL

    init_local_variables();
}

/* ---------------------------------------------------------------------------------------- */

EXP_Class::~EXP_Class( ){
    delete param;
    //delete[] str_cylinder;
    //delete[] str_light;
}

/* ---------------------------------------------------------------------------------------- */

void EXP_Class::init_local_variables( void ){
    //In this function you have to create your objects like agents,
    //sensors, obstables world characteristics etc.
    eval       = 0; //This is the counter for the evaluation
    iter       = 0; //This is the counter for the iterations
    genotype   = 0; //This is the counter for the solutions of genotypes
    if( evolution ) param->dump_simulation_seed ( );//This is to dump the seed number in EXP/DATA

    //This is to init the network controller
    param->init_controllers( );
    param->init_objects( );
    param->init_agents( );
    centre.assign(3,0.0);
    //init and reset the array to store Infra-red sensor readings
    ir_readings.assign( param->agent[0]->num_IR_sensors, 0.0);
    camera_sector_readings.assign( param->agent[0]->num_camera_sectors, 0.0);
    bearing_sector_reading.assign( param->agent[0]->num_bearing_sectors,0.0);
    mouse_reading.assign(param->agent[0]->num_mouse_parameters,0.0);
    agent_interface.resize( param->num_agents );
    for(int r = 0; r < param->num_agents; r++){
        agent_interface[r].inputs.assign  ( param->nets[r]->get_num_input(), 0.0);
        agent_interface[r].outputs.assign ( param->nets[r]->get_num_output(), 0.0);
    }
    REEVAL_FITNESS.resize(5);
    for( int e=0;e< 5; e++)
        REEVAL_FITNESS[e].assign (12,0.0);
    DIST_FITNESS.resize(5);
    for( int e=0;e< 5; e++)
        DIST_FITNESS[e].assign (12,0.0);
    pos.resize(param->num_agents);
    rot.resize(param->num_agents);
    for(int i=0;i<param->num_agents;i++){
        pos[i].assign(3,0.0);
        rot[i].assign(3,0.0);
    }
    this->set_agent_position();
}

/* ---------------------------------------------------------------------------------------- */

void EXP_Class::init_evolutionary_run( void ){
    //max_distance=100.0;
    if(evolution){
        generation = 0; //This is the counter for the generation
        param->init_ga( param->nets[0]->get_genotype_length(), 1 );
    }
    else if(re_evaluation){
        generation = from_gen;
    }

}

/* ---------------------------------------------------------------------------------------- */

void EXP_Class::init_genotype_loop( ){ // this loop for each population
    genotype   = 0;//This is the counter for the population or solutions (genotype)
    generational_avg_fitness     = 0.0;
    generational_max_fitness     = 0.0;
    generational_min_fitness     = 0.0;
    if(re_evaluation){
        sprintf(param->genomeFileName_for_reval, "%s%s_pop_G%d.geno", param->genomeFileDir, param->genomeFileName, generation );
        from_genome_to_controllers(param->genomeFileName_for_reval, genotype);
    }
}


/* ---------------------------------------------------------------------------------------- */

void EXP_Class::set_agent_position(){

    param->resetPhysicsState();
    for(int obj=0;obj<param->num_objects;obj++){
        if(param->object[obj]->get_mass() != 0.0){
            param->object[obj]->reset_pos();
            param->object[obj]->body[param->object[obj]->object_id]->clearForces();
            param->object[obj]->body[param->object[obj]->object_id]->setLinearVelocity(btVector3(0.0,0.0,0.0));
            param->object[obj]->body[param->object[obj]->object_id]->setAngularVelocity(btVector3(0.0,0.0,0.0));
        }
    }

    //  if(eval < 2)
    if(eval == 0){
        for(int i=4;i<16;i++)
            if(!param->agent[i]->is_removed())
                param->remove_agent(param->agent[i]);
        param->object[0]->set_mass(0.630);
    }
    /* }else if(eval < 4){
        if(eval == 2){
            for(int i=2;i<4;i++)
                if(param->agent[i]->is_removed())
                    param->add_agent(param->agent[i]);
            param->object[0]->set_mass(0.530);
            param->num_iterations = 550;
        }
    }else if(eval < 6){
        if(eval == 4){
            for(int i=4;i<6;i++)
                if(param->agent[i]->is_removed())
                    param->add_agent(param->agent[i]);
            param->object[0]->set_mass(0.700);
            param->num_iterations = 60;
        }
    }  */



    //    if(eval%2){

    //        pos[0][0] = (gsl_rng_uniform_pos( GSL_randon_generator::r_rand )*0.50 - 0.25);
    //        pos[0][2] = sqrt(0.0625 - pos[0][0]*pos[0][0]);


    //        pos[1][0] = pos[0][0] + 0.10;
    //        pos[3][0] = pos[0][0] - 0.10;
    //        pos[5][0] = pos[0][0] - 0.20;
    //        pos[7][0] = pos[0][0] - 0.30;
    //        pos[2][0] = pos[0][0] + 0.20;
    //        pos[4][0] = pos[0][0] + 0.30;
    //        pos[6][0] = pos[0][0] + 0.40;
    //        pos[8][0] = pos[0][0] - 0.40;
    //        pos[9][0] = pos[0][0] + 0.50;
    //        pos[10][0] = pos[0][0] - 0.50;
    //        pos[11][0] = pos[0][0] + 0.60;
    //        pos[12][0] = pos[0][0] + 0.10;
    //        pos[13][0] = pos[0][0] - 0.10;
    //        pos[14][0] = pos[0][0] - 0.20;
    //        pos[15][0] = pos[0][0] - 0.30;
    //        pos[1][2] = pos[2][2] = pos[3][2] =   pos[4][2] = pos[5][2] = pos[6][2] = pos[7][2] =  pos[8][2] = pos[9][2] = pos[10][2] = pos[11][2] = pos[0][2];
    //        pos[12][2] = pos[13][2] = pos[14][2] =   pos[15][2] = pos[5][2] - 0.15;
    //        for(int i=0;i<param->num_agents;i++){
    //            rot[i][1] = atan2(pos[i][2], pos[i][0]);
    //            if(rot[i][1] < 0.0) rot[i][1] = ( TWO_PI + rot[i][1]);
    //            rot[i][1] = (TWO_PI - rot[i][1]) + PI + gsl_rng_uniform_pos( GSL_randon_generator::r_rand )*PI/2 - (PI/4);
    //        }

    //    }else{
    //do{
//        pos[0][0] = (gsl_rng_uniform_pos( GSL_randon_generator::r_rand )*1.00 - 0.50);
//        pos[0][2] = sqrt(0.25 - pos[0][0]*pos[0][0]);

//        pos[1][0] = pos[3][0] = pos[8][0] = pos[9][0] = pos[15][0] = -pos[0][0];
//        pos[5][0] = pos[3][0] + 0.10;
//        pos[2][0] = pos[12][0] = pos[13][0] = pos[14][0] = pos[0][0];
//        pos[4][0] = pos[0][0] + 0.10;
//        pos[1][2] = pos[2][2] = -pos[0][2];
//        pos[4][2] = pos[2][2] + 0.10;
//        pos[6][2] = pos[2][2] + 0.20;
//        pos[3][2] = pos[0][2];
//        pos[5][2] = pos[0][2] - 0.10;
//        pos[7][2] = pos[0][2] - 0.20;
//        pos[8][2] = pos[0][2] + 0.30;
//        pos[9][2] = pos[2][2] - 0.30;
//        pos[10][2] = pos[0][2] + 0.20;
//        pos[11][2] = pos[0][2] - 0.30;
//        pos[12][2] = pos[0][2] + 0.40;
//        pos[13][2] = pos[1][2] - 0.40;
//        pos[14][2] = pos[1][2] + 0.40;
//        pos[15][2] = pos[1][2] - 0.40;

    double min_dist = 0.10;//(1.5 + (2.0 * m_agents[0]->get_radius())) * (1.5 + (2.0 * m_agents[0]->get_radius()));
    bool flag = false;
    double angle = 0.0;
    double dist = 0.0;

    std::vector < double > sector;
    sector.resize(param->num_active_agents);
    for (int i = 0; i < sector.size(); i++) sector[i] = 2.0*PI/((double)param->num_active_agents) - (2.0*PI/((double)param->num_active_agents)*i);

    for(int r = 0; r < param->num_active_agents; r++){
        do{
            flag = false;
            int pos1 = gsl_rng_uniform_int (GSL_randon_generator::r_rand, sector.size() );
            double index = sector[pos1];
            sector.erase( sector.begin() + pos1 );
            if( sector.empty() ) {
                sector.resize(param->num_active_agents);
                for (int i = 0; i < sector.size(); i++) sector[i] = 2.0*PI/((double)param->num_active_agents) - (2.0*PI/((double)param->num_active_agents)*i);;
            }
            angle = index - (gsl_rng_uniform_pos(GSL_randon_generator::r_rand) * PI/3.0);
            //dist = ((floor_target_dim[nest_id][1] - 5.0) * gsl_rng_uniform_pos(GSL_randon_generator::r_rand));
            //cerr << " pos = " << pos << " sector = " << sector[pos] << " index = " << index << " angle = " << ((angle*180.0)/PI) << endl;
            //getchar();

            pos[r][0]=   cos(angle) * 0.50;
            pos[r][2] =  sin(angle) * 0.50;
            for(int ar = 0; ar < r; ar++){
                double x_off = pos[r][0] - pos[ar][0];
                double y_off = pos[r][2] - pos[ar][2];
                if( ((x_off*x_off)+(y_off*y_off)) < (min_dist*min_dist) ){
                    flag = true;
                    break;
                }
            }
        } while( flag );

    }

    //}while(!(find_distance(pos[0],pos[3])>0.10 && find_distance(pos[1],pos[2])>0.10));

    for(int i=0;i<param->num_agents;i++){
        rot[i][1] = atan2(pos[i][2], pos[i][0]);
        if(rot[i][1] < 0.0) rot[i][1] = ( TWO_PI + rot[i][1]);
        rot[i][1] = (TWO_PI - rot[i][1]) + PI + gsl_rng_uniform_pos( GSL_randon_generator::r_rand )*PI/2 - (PI/4);
    }


//                pos[0][2] = 0.065;
//                pos[0][0] = 0.0;
//                rot[0][1] = 0.5*PI;
    //            pos[1][2] = 0.62;
    //            pos[1][0] = 0.50;
    //            rot[1][1] = 0.5 * PI;
    //            pos[2][2] = 0.34;
    //            pos[2][0] = 0.50;
    //            rot[2][1] = PI;

    for(int r=0;r<param->num_agents;r++)
        if(!param->agent[r]->is_removed())
            param->agent[r]->set_robot_pos_rot( pos[r], rot[r] );

}

/* ---------------------------------------------------------------------------------------- */

void EXP_Class::init_evaluations_loop( ){ // this loop for agents life time for single trail
    //printf(" individual = %d\n",genotype);
    eval = 0;//This is the counter for the evalutions
    counter = 0;
    rank = 0;
    success = 0;
    fitnesses_mult.assign(675,0.0);
    FINAL_FITNESS.assign(param->num_evaluations,0.0);
    REEVAL_FITNESS.resize(5);
    for( int e=0;e< 5; e++)
        REEVAL_FITNESS[e].assign (12,0.0);
    DIST_FITNESS.resize(5);
    for( int e=0;e< 5; e++)
        DIST_FITNESS[e].assign (12,0.0);
    evaluatoinal_min_fitness.assign(5,1.0);
    num_collision = 0;
    collision_ratio = 0.0;
    param->reset_seed( );
    //start_distance=find_distance();

    //There are various thing you want to do here
    // eg: resettng the fitness to zero (const vector <double> & FINAL_FITNESS)
    // init the agent position, etc
    //printf("set the initial agent position and fitness to zero ...\n");
    //  set_agent_position();
    if( evolution ) {

        for(int r=0; r < param->num_agents; r++) {
#ifdef _HETERO_GROUP_SELECTION_
            //This is where we create a network from solution num: genotype
            param->nets[r]->init(  param->ga->get_solution( genotype, r ) );
#else
            //This is where we create a network from solution num: genotype
            param->nets[r]->init(  param->ga->get_solution( genotype, 0 ) );
#endif

        }
    }
}

/* ---------------------------------------------------------------------------------------- */

void EXP_Class::init_single_evaluation( void ){ 
    iter = 0;//This is the counter for the iteration mean for starting trail
    remaining_iter = 0;
    init_agent_dist_to_object = 0.30;
    init_object_dist_to_nest = 0.62;
    transported = false;
    dist_achieved = 0.0;
    agents_closeness_to_object = 0.0;
    object_closeness_to_nest = 0.0;
    set_agent_position();
    for(int r=0; r < param->num_agents; r++) {
        param->nets[r]->reset(); //Maybe

    }
    //printf("\n  eval%d internal_fitness = %f",eval, internal_fitness);

}

/* ---------------------------------------------------------------------------------------- */

void EXP_Class::from_genome_to_controllers( const char *str_source, const int which_genotype ){
    int gen_length = 0;
    vector <chromosome_type> genes;
    vector <chromosome_type> copy_genes;

#ifdef _HETERO_GROUP_SELECTION_
    upload_parameters_from_file <chromosome_type> ( str_source, which_genotype, genes, &gen_length, param->num_agents );
    copy_genes.assignagent_interface[r].( gen_length*param->num_agents, 0.0);
#else
    upload_parameters_from_file <chromosome_type> ( str_source, which_genotype, genes, &gen_length, 1 );
    copy_genes.assign( gen_length, 0.0);
#endif

    for(int r=0; r < param->num_agents; r++) {
        param->nets[r]->set_genotype_length ( copy_genes.size() );
#ifdef _HETERO_GROUP_SELECTION_
        for(int g = 0; g < copy_genes.size(); g++){
            copy_genes[g] = genes[g+(copy_genes.size()*r)];
        }
        param->nets[r]->init(  copy_genes );max_geno_to_eval
        #else
        param->nets[r]->init( genes );
#endif
    }
}

/* ---------------------------------------------------------------------------------------- */

void EXP_Class::adv ( void ){
    if( viewing ) stop_iterations_loop( );
    update_sensors( );
    update_controllers ( );
    update_Actuators();
    for(int i=0;i < 6;i++){
          update_world();
          param->world->stepSimulation( param->physics_step);
    }
    //compute_internal_fitness();
    manage_collisions ();
    if(find_distance(param->object[0]->get_pos(),param->object[0]->start_pos) > 2.0){
          remaining_iter = param->num_iterations - iter;
          iter = param->num_iterations;
    }
    ++iter;
    printf("\n iter = %d",iter);

}


/* ---------------------------------------------------------------------------------------- */

void EXP_Class::update_sensors( void ){

    for(int r=0; r < param->num_agents; r++){
        if(!param->agent[r]->is_removed()){
            for (int i = 0; i < ir_readings.size(); i++) ir_readings[i] = -1.0;
            for (int i = 0; i < camera_sector_readings.size(); i++) camera_sector_readings[i] = 0.0;
            //   for (int i = 0; i < bearing_sector_reading.size(); i++) bearing_sector_reading[i] = 0.0;

            param->agent[r]->get_IR_reading(ir_readings);
            param->agent[r]->add_noise(ir_readings);

            for( int i = 0; i < param->agent[r]->num_IR_sensors; i++)
                agent_interface[r].inputs[i] = ir_readings[i];


            param->agent[r]->get_camera_reading(camera_sector_readings);
            for( int i = 0; i < param->agent[r]->num_camera_sectors; i++){
                agent_interface[r].inputs[i+8] = camera_sector_readings[i];
        }

        }
    }
}



/* ---------------------------------------------------------------------------------------- */


void EXP_Class::update_controllers( void ){

    for(int r=0; r < param->num_agents; r++){
        if(!param->agent[r]->is_removed()){
            param->nets[r]->step( agent_interface[r].inputs /* const vector <double> &inputs*/ , agent_interface[r].outputs /* vector <double> &outputs */ );
        }
    }

}

/* ---------------------------------------------------------------------------------------- */

void EXP_Class::update_Actuators( void ){
    vector <double> outputs;
    outputs.resize(4);
    outputs[0] = 1.0; // right wheel
    outputs[1] = 0.0;
    outputs[2] = 1.0; //left wheel
    outputs[3] = 0.0;


    for(int r=0; r < param->num_agents; r++){
        //update robot wheels velocity
        if(!param->agent[r]->is_removed()){
            param->agent[r]->set_vel(agent_interface[r].outputs);
        }
    }
}

/* ---------------------------------------------------------------------------------------- */


/* ---------------------------------------------------------------------------------------- */

void EXP_Class::update_world( void ){

    for(int r=0; r < param->num_agents; r++){
        if(!param->agent[r]->is_removed()){
            //update robot position and rotation
            param->agent[r]->update_pos_rot( );
        }
    }
}

/* ---------------------------------------------------------------------------------------- */

void EXP_Class::manage_collisions (void ){


    for(int r = 0;r<param->num_agents;r++){
        if(!param->agent[r]->is_removed()){

            for (int i = 0; i < mouse_reading.size(); i++) mouse_reading[i] = 0.0;

            param->agent[r]->get_mouse_reading(mouse_reading);
            //            for(int i=0;i<mouse_reading.size();i++)
            //                printf("Rob%d mouse%d = %f ",r,i,mouse_reading[i]);
            //            printf("\n");
            //            for(int i=0;i<agent_interface[r].outputs.size();i++)
            //                printf("Rob%d out%d = %f ",r,i,agent_interface[r].outputs[i]);
            //            printf("\n=====================================================================================================\n");
            agent_interface[r].inputs[11] = agent_interface[r].outputs[0];
            agent_interface[r].inputs[12] = agent_interface[r].outputs[1];
            agent_interface[r].inputs[13] = agent_interface[r].outputs[2];
            agent_interface[r].inputs[14] = agent_interface[r].outputs[3];
            agent_interface[r].inputs[15] = mouse_reading[0];
            agent_interface[r].inputs[16] = mouse_reading[1];
            agent_interface[r].inputs[17] = mouse_reading[2];
            agent_interface[r].inputs[18] = mouse_reading[3];

            //            if(!param->object[1]->is_removed()){
            //                while(param->object[1]->collision_counter){
            //                    if(param->object[1]->collision_objects[(param->object[1]->collision_counter)-1][1] == BRICK){
            //                        iter = param->num_iterations;
            //                    }
            //                    delete param->agent[r]->collision_objects[(param->object[1]->collision_counter)-1];
            //                    param->object[1]->collision_counter--;
            //                }
            //            }
        }
    }

}


/* ---------------------------------------------------------------------------------------- */

void EXP_Class::compute_internal_fitness(){

    for(int r = 0;r<param->num_agents;r++){
        if(!param->agent[r]->is_removed()){
            current_dist =find_distance(param->object[0]->get_pos(), param->agent[r]->get_pos());
            if(current_dist > init_agent_dist_to_object){
                init_agent_dist_to_object = current_dist;
            }
        }
    }

}

/* ---------------------------------------------------------------------------------------- */

void EXP_Class::compute_fitness( void ){


    AGGREGATE_FITNESS = 0.0;
    for(int r = 0;r<param->num_agents;r++){
        if(!param->agent[r]->is_removed()){
            current_dist =find_distance(param->object[0]->get_pos(), param->agent[r]->get_pos());
            if(current_dist < 0.18) current_dist = 0.0;
            AGGREGATE_FITNESS += (init_agent_dist_to_object - current_dist)/init_agent_dist_to_object;
        }
    }

    agents_closeness_to_object = AGGREGATE_FITNESS / param->num_active_agents;

    // printf("\n agents_closeness_to_object = %f ", agents_closeness_to_object);
    // printf("\n");
    printf("\n distance = %f",find_distance(param->object[0]->get_pos(),param->object[0]->start_pos));
    printf("\n");
    FINAL_FITNESS[0] = (find_distance(param->object[0]->get_pos(),param->object[0]->start_pos) * 10) + agents_closeness_to_object + (remaining_iter * 0.0267);
    //printf("\n external fitness eval%d = %lf ",eval, FINAL_FITNESS[0] );


}



/* ---------------------------------------------------------------------------------------- */

double EXP_Class::find_distance(const vector<double> &_pos1, const vector<double> &_pos2 ){

    return sqrt(((_pos2[0]-_pos1[0])*(_pos2[0]-_pos1[0]) + (_pos2[2]-_pos1[2])*(_pos2[2]-_pos1[2])));

}


/* ---------------------------------------------------------------------------------------- */

void EXP_Class::finalise_single_evaluation ( void ){

    //printf(" compute fitness for eval%d \n",eval);
    // printf(" evaluate compute_fitness = %lf \n",param->agent[0]->get_robot_pos( )[2]);
    compute_fitness();
    //  internal_fitness /= param->num_iterations;
    //  //printf("\n internal fitness =%f",internal_fitness);
    //  FINAL_FITNESS[0] += internal_fitness;
    //  //printf("\n final fitness =%f",FINAL_FITNESS[0]);
    //  FINAL_FITNESS[0] = (FINAL_FITNESS[0]/60.0) + 1.0;
    //  fitnesses_mult *= FINAL_FITNESS[0];

}

/* ---------------------------------ir_readings------------------------------------------------------- */

void EXP_Class::finalise_evaluations_loop( void ){

    generational_avg_fitness += FINAL_FITNESS[0] /*FINAL_FITNESS*/;
    if( genotype == 0 ){
        generational_max_fitness     = FINAL_FITNESS[0]/*FINAL_FITNESS*/;
        generational_min_fitness     = FINAL_FITNESS[0] /*FINAL_FITNESS*/;
    }
    else{
        if( /*FINAL_FITNESS*/ FINAL_FITNESS[0] > generational_max_fitness ){
            generational_max_fitness     = FINAL_FITNESS[0] /*FINAL_FITNESS*/;
        }
        else if(/*FINAL_FITNESS*/FINAL_FITNESS[0]  < generational_min_fitness )
            generational_min_fitness = FINAL_FITNESS[0]/*FINAL_FITNESS*/;
    }

}

/* ---------------------------------------------------------------------------------------- */

void EXP_Class::assign_fitness_to_genotype( void ){

    //cerr << " ind = " << genotype << " fit = " << group[0].total_fitness[0] << endl;
    param->ga->assign_fitness( genotype,  FINAL_FITNESS /* const vector <double> & FINAL_FITNESS */ );
}

/* ---------------------------------------------------------------------------------------- */

void EXP_Class::finalise_genotypes_loop( void ){
    if( !(generation%param->dump_stats_every) )
        dump_statistics( param->statsFileDir, param->statsFileName, generation);
    if( !(generation%param->dump_genome_every) )
        param->ga->dump_genome_into_file( param->genomeFileDir,  param->genomeFileName, generation );

    //This is the function that create s a new generation of solutions (genotypes)
    param->ga->breeding();

}

/* ---------------------------------------------------------------------------------------- */

void EXP_Class::finalise_evolutionary_run( void ){}

/* ---------------------------------------------------------------------------------------- */

bool EXP_Class::stop_iterations_loop( void ){
    //This function should not be changed
    if(iter >= param->num_iterations){
        finalise_single_evaluation( );
        iter = 0;
        if( viewing ) stop_evaluations_loop( );
        return false;
    }
    else
        return true;
}

/* ---------------------------------------------------------------------------------------- */

bool EXP_Class::stop_evaluations_loop( void ){

    eval++;
    if( evolution ){
        if( eval >= param->num_evaluations ) {
            eval = 0;
            finalise_evaluations_loop( );
            assign_fitness_to_genotype();
            return false;
        }
        else{
            return true;
        }
    }
    else{
        if( viewing ) {
            if( eval >= param->num_evaluations ) {
                eval = 0;
                finalise_evaluations_loop( );
            }
            init_single_evaluation( );
            return true;
        }

        else if( re_evaluation ){
            if( eval >= param->num_evaluations ) {
                eval = 0;
                finalise_evaluations_loop( );
                return false;
            }
            else{
                return true;
            }
        }

    }
}

/* ---------------------------------------------------------------------------------------- */

bool EXP_Class::stop_genotypes_loop( void ){
    if(evolution){
        genotype++;
        if( genotype >= param->ga->get_num_tot_solutions() ) {
            finalise_genotypes_loop( );
            genotype = 0;
            return false;
        }
        else
            return true;
    }
    else if(re_evaluation){

        genotype = 0;
        return false;
    }
}

/* ---------------------------------------------------------------------------------------- */

bool EXP_Class::stop_generations_loop( void ){
    generation++;
    if(evolution){
        if( generation >= param->num_generations ) {
            finalise_evolutionary_run();
            generation = 0;
            return false;
        }
        else
            return true;
    }
    else if(re_evaluation){
        if( generation > to_gen ){
            return false;
        }
        else{
            return true;
        }

    }
}

/* ---------------------------------------------------------------------------------------- */

void EXP_Class::dump_statistics( const char *locationOfFileTodump, 
                                 const char *fileName, int generation ) {
    char fname_statistics[500];
    sprintf(fname_statistics, "%s%s.data", locationOfFileTodump, fileName );
    ofstream outfile;
    outfile.open (fname_statistics, ios::app);
    outfile.setf(ios::fixed);

    outfile << " " << generation
            << " " << generational_max_fitness                                       //the fitness of the best
            << " " << (generational_avg_fitness)/(double)(param->ga->get_num_tot_solutions() )  // the mean fitness
            << " " << generational_min_fitness                                       //the fitness of the worst
            << endl;
    outfile.close();
}

/* ---------------------------------------------------------------------------------------- */


void EXP_Class::dump_reevaluation_results( const char *locationOfFileTodump, const char *fileName, int generation ) {
    char fname_statistics[500];
    sprintf(fname_statistics, "%s%s.results", locationOfFileTodump, fileName );
    ofstream outfile;
    outfile.open (fname_statistics, ios::app);
    outfile.setf(ios::fixed);
    outfile << " " << generation;
    for(int i=0;i<5;i++){
        outfile << " " <<  evaluatoinal_min_fitness[i];
    }
    outfile << endl;
    outfile.close();
}

/* ---------------------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------------------------ */

void EXP_Class::dump_reevaluation_raw_data( const char *locationOfFileTodump, const char *fileName, int generation ) {
    char fname_statistics[500];
    sprintf(fname_statistics, "%s%s.rawdata", locationOfFileTodump, fileName );
    ofstream outfile;
    outfile.open (fname_statistics, ios::app);
    outfile.setf(ios::fixed);
    outfile << " " << generation;
    for(int i=0;i<5;i++){
        for(int j=0;j<12;j++){
            outfile << " " <<  REEVAL_FITNESS[i][j];
        }
    }
    outfile << endl;
    outfile.close();
}

/* ---------------------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------------------------ */

void EXP_Class::dump_reevaluation_distance( const char *locationOfFileTodump, const char *fileName, int generation ) {
    char fname_statistics[500];
    sprintf(fname_statistics, "%s%s.dist", locationOfFileTodump, fileName );
    ofstream outfile;
    outfile.open (fname_statistics, ios::app);
    outfile.setf(ios::fixed);
    outfile << " " << generation;
    for(int i=0;i<5;i++){
        for(int j=0;j<12;j++){
            outfile << " " <<  DIST_FITNESS[i][j];
        }
    }
    outfile << endl;
    outfile.close();
}

/* ---------------------------------------------------------------------------------------- */
/* ---------------------------------------------------------------------------------------- */
/* ---------------------------------------------------------------------------------------- */




