#include "parameters.h"

bool callbackFunc(btManifoldPoint& cp, const btCollisionObjectWrapper* obj1, int id1, int index1, const btCollisionObjectWrapper* obj2, int id2, int index2);

//void myTickCallback(btDynamicsWorld *world, btScalar timeStep);

/* ---------------------------------------- */

void Parameters::read_run_parameter_file( const char *run_name ){

    ifstream I ("../MISC/init_run.txt");
    if(!I) {
        cerr << "File for Run Parameters not found" <<endl;
        exit(0);
    }

    /* ------------------------------- */
    //Load parameters from init_run.txt
    /* ------------------------------- */
    root_seed         = getInt('=', I);
    num_generations   = getInt('=', I);
    num_evaluations   = getInt('=', I);
    num_iterations    = getInt('=', I);
    num_agents        = num_active_agents        = getInt('=', I);

    dump_stats_every  = getInt('=', I);
    dump_genome_every = getInt('=', I);

    getStr('=', I, statsFileDir );
    getStr('=', I, revalFileDir );
    getStr('=', I, genomeFileDir );
    getStr('=', I, type_of_controller );
    getStr('=', I, type_of_ga );

    simulation_time_step = getDouble('=', I);
    physics_step = simulation_time_step/6.0;

    num_bricks        = getInt('=', I);
    num_cylinders     = getInt('=', I);
    num_spheres       = getInt('=', I);
    num_lights        = getInt('=', I);
    num_objects       = num_bricks + num_cylinders +  num_spheres + num_lights;

    int num_obj_properties = getInt('=', I);

    if ( num_objects ){
        objects_data = new double*[num_objects];

        for(int b=0; b < num_objects; b++){
            objects_data[b] = new double[num_obj_properties];
        }

        for(int c=0; c < num_obj_properties; c++) {
            for(int b=0; b < num_objects; b++) {
                objects_data[b][c] = getDouble('=', I);
            }
        }
    }
    /* ------------------------------- */
    //End of loading parameters
    /* ------------------------------- */

    strcpy(statsFileName, run_name);
    strcpy(revalFileName, run_name);
    strcpy(rootSeedFileDir, statsFileDir );
    strcpy(rootSeedFileName, run_name );
    strcpy(genomeFileName, run_name);
}


/* ---------------------------------------- */

Parameters::Parameters(const char *_run_name_ ){
    root_seed = 0;
    count = 0;
    run_name = new char[500];
    collision_peer = new int*[100];
    sprintf(run_name, "%s", _run_name_);
    read_run_parameter_file( run_name );//Open the first configuration file called init_run.txt
    init_physics_param( );
}


/* ---------------------------------------- */

void Parameters::setDeltaTime(btScalar deltaTime)
{
    /*
    double ms = deltaTime*1000000.;
    double minFPS = 1000000.f/60.f;
    if (ms > minFPS)
        ms = minFPS;

    m_Time += ms;

*/
}

/* ---------------------------------------- */


btScalar Parameters::getDeltaTimeMicroseconds()
{
    /*
#ifdef USE_BT_CLOCK
        btScalar dt = (btScalar)m_clock.getTimeMicroseconds();
        m_clock.reset();
        return dt;
#else
        return btScalar(16666.);
#endif
*/
}
/* ---------------------------------------- */

void Parameters::myProcessCallback(btScalar timeStep) {
    /*
    for(int i=0;i<num_agents;i++){
    agent[i]->body->clearForces();
    agent[i]->left_wheel->clearForces();
    agent[i]->right_wheel->clearForces();
    agent[i]->body->applyGravity();
    agent[i]->right_wheel->applyGravity();
    agent[i]->left_wheel->applyGravity();
    agent[i]->right_hinge->setLimit(0.0, 0.0);
    agent[i]->right_hinge->enableAngularMotor(false, 0,0);
    agent[i]->left_hinge->setLimit(0.0, 0.0);
    agent[i]->left_hinge->enableAngularMotor(false, 0,0);
    agent[i]->body->setActivationState(0);
    agent[i]->right_wheel->setActivationState(0);
    agent[i]->left_wheel->setActivationState(0);
    agent[i]->right_hinge->setMotorTarget(0.0, timeStep);
    agent[i]->left_hinge->setMotorTarget(0.0,timeStep);
    }
*/
}

/* ---------------------------------------- */
/*
void myTickCallback(btDynamicsWorld *world, btScalar timeStep) {
    Parameters *w = static_cast<Parameters *>(world->getWorldUserInfo());
    w->myProcessCallback(timeStep);
}
*/

/* ---------------------------------------- */

void Parameters:: init_physics_param( void ){

    this->collisionConfig=new btDefaultCollisionConfiguration();
    this->dispatcher=new btCollisionDispatcher(collisionConfig);
    btVector3 worldMin(-5,-1,-5);
    btVector3 worldMax(5,1,5);
    this->broadphase = new btAxisSweep3(worldMin,worldMax);
    this->solver=new btSequentialImpulseConstraintSolver();
    this->world=new btDiscreteDynamicsWorld(dispatcher,broadphase,solver,collisionConfig);
    this->world->setGravity(btVector3(0.0,-9.81,0.0));  //gravity on Earth
    this->world->getSolverInfo().m_numIterations = 110;
    world->getSolverInfo().m_solverMode |= SOLVER_USE_2_FRICTION_DIRECTIONS  | SOLVER_SIMD;// | SOLVER_RANDMIZE_ORDER;
    //gContactAddedCallback = callbackFunc;



}

/* ---------------------------------------------------------------------------------------- */



/* ---------------------------------------------------------------------------------------- */

Parameters::~Parameters(){

}



/* ------------------------------------------------------------------------ */



void	Parameters::resetPhysicsState()
{


    int numObjects = 0;
    int i;

    if (world)
    {
        int numConstraints = world->getNumConstraints();
        for (i=0;i<numConstraints;i++)
        {
            world->getConstraint(0)->setEnabled(true);
        }
        numObjects = world->getNumCollisionObjects();

        ///create a copy of the array, not a reference!
        btCollisionObjectArray copyArray = world->getCollisionObjectArray();




        for (i=0;i<numObjects;i++)
        {
            btCollisionObject* colObj = copyArray[i];
            btRigidBody* body = btRigidBody::upcast(colObj);
            if (body)
            {
                if (body->getMotionState())
                {
                    btDefaultMotionState* myMotionState = (btDefaultMotionState*)body->getMotionState();
                    myMotionState->m_graphicsWorldTrans = myMotionState->m_startWorldTrans;
                    body->setCenterOfMassTransform( myMotionState->m_graphicsWorldTrans );
                    colObj->setInterpolationWorldTransform( myMotionState->m_startWorldTrans );
                    colObj->forceActivationState(ACTIVE_TAG);
                    colObj->activate();
                    colObj->setDeactivationTime(0);
                    //colObj->setActivationState(WANTS_DEACTIVATION);
                }
                //removed cached contact points (this is not necessary if all objects have been removed from the dynamics world)
                if (world->getBroadphase()->getOverlappingPairCache())
                    world->getBroadphase()->getOverlappingPairCache()->cleanProxyFromPairs(colObj->getBroadphaseHandle(),dispatcher);

                btRigidBody* body = btRigidBody::upcast(colObj);
                if (body && !body->isStaticObject())
                {
                    btRigidBody::upcast(colObj)->setLinearVelocity(btVector3(0,0,0));
                    btRigidBody::upcast(colObj)->setAngularVelocity(btVector3(0,0,0));
                }
            }

        }

        ///reset some internal cached data in the broadphase
        world->getBroadphase()->resetPool(dispatcher);
        world->getConstraintSolver()->reset();

    }

}


/* ------------------------------------------------------------------------ */


bool callbackFunc(btManifoldPoint& cp, const btCollisionObjectWrapper* obj1, int id1, int index1, const btCollisionObjectWrapper* obj2, int id2, int index2)
{
    World_Entity* object1 = (World_Entity*)obj1->getCollisionObject()->getUserPointer();
    World_Entity* object2 = (World_Entity*)obj2->getCollisionObject()->getUserPointer();
    if(object1->get_type_id() != 20){
        if(object1->get_type_id() != ROBOT){
            if(object2->get_type_id()== CYLINDER){
                object2->collision_objects[object2->collision_counter] = new int[2];
                object2->collision_objects[object2->collision_counter][0] = object1->get_index();
                object2->collision_objects[object2->collision_counter][1] = object1->get_type_id();
                object2->collision_counter++;
            }else{
                object1->collision_objects[object1->collision_counter] = new int[2];
                object1->collision_objects[object1->collision_counter][0] = object2->get_index();
                object1->collision_objects[object1->collision_counter][1] = object2->get_type_id();
                object1->collision_counter++;
            }
        }
    }

    return false;
}


/* ---------------------------------------- */

void Parameters::add_agent(SIMPLE_Agents* agent){

    vector <double> pos;
    pos.assign(3,0.0);
    vector <double> rot;
    rot.assign(3,0.0);
    pos[0] = 4.80;
    pos[2] = 4.80;
    agent->set_robot_pos_rot( pos, rot );
    agent->right_hinge->setLimit(1,-1,1.0,0.3,1);
    agent->left_hinge->setLimit(1,-1,1.0,0.3,1);
    agent->body->setActivationState(1);
    agent->right_wheel->setActivationState(1);
    agent->left_wheel->setActivationState(1);
    agent->set_removed(false);
    num_active_agents++;
}

/* ---------------------------------------- */

void  Parameters::remove_agent(SIMPLE_Agents *agent){
    vector <double> pos;
    pos.assign(3,0.0);
    vector <double> rot;
    rot.assign(3,0.0);
    pos[0] = 5.8;
    pos[2] = 5.8;
    agent->set_robot_pos_rot( pos, rot );
    agent->right_hinge->setLimit(0.0, 0.0);
    agent->left_hinge->setLimit(0.0, 0.0);
    agent->body->setActivationState(0);
    agent->right_wheel->setActivationState(0);
    agent->left_wheel->setActivationState(0);
    agent->set_removed(true);
    num_active_agents--;

}

/* ---------------------------------------- */

void Parameters::init_random_generator( void ){
    GSL_randon_generator::init_generator( root_seed );
}

/* ---------------------------------------- */

void Parameters::init_objects( ){
    plane = new SIMPLE_Plane( world );
    plane->body[0]->setUserPointer(plane);

    if( num_objects ){


        object = new SIMPLE_Objects*[num_objects];

        int count = 0, ob = 0;

        while( count < num_bricks && ob < num_objects ) {

            object[ob] = new SIMPLE_Brick( ob, objects_data[ob], world);
            //body->setUserPointer(bodies[bodies.size()-1]);
            object[ob]->body[0]->setUserPointer(object[ob]);
            //object[ob]->body[1]->setUserPointer(object[ob]);
            //object[ob]->body[2]->setUserPointer(object[ob]);
            object[ob]->set_crashed(false);
            ob++;
            count++;
        }
        count = 0;
        while( count < num_cylinders && ob < num_objects ) {
            object[ob] = new SIMPLE_Cylinder ( ob, objects_data[ob], world);
            object[ob]->body[0]->setUserPointer(object[ob]);
            ob++;
            count++;
        }
        count = 0;
        while( count < num_spheres && ob < num_objects ) {
            object[ob] = new SIMPLE_Sphere ( ob, objects_data[ob], world);
            object[ob]->body[0]->setUserPointer(object[ob]);
            ob++;
            count++;
        }
    }
}

/* ---------------------------------------- */

void Parameters::init_agents( void ){
    vector <double> colour;
    colour.assign(3, 1.0);
    agent = new SIMPLE_Agents*[num_agents];
    for (int r = 0; r < num_agents; r++){
        agent[r] = new SIMPLE_Agents( r, simulation_time_step, world);
        agent[r]->set_crashed(false);
        agent[r]->body->setUserPointer(agent[r]);
        agent[r]->left_wheel->setUserPointer(agent[r]);
        agent[r]->right_wheel->setUserPointer(agent[r]);
        colour[0] = 1.0;
        colour[1] = 0.0;
        colour[2] = 0.0;
        agent[r]->set_colour( colour );
    }

}


/* ---------------------------------------- */

void Parameters::init_controllers( void ){
    nets = new Controller* [num_agents];
    if ( !strcmp ( type_of_controller, "ctrnn_three_layers" ) ){
        for(int r=0; r < num_agents; r++) nets[r] = new Ctrnn3Layers(  );
    }
    else if ( !strcmp ( type_of_controller, "ctrnn_three_layers_hebb" ) ){
        printf("create controller as number of agents....\n");
        for(int r=0; r < num_agents; r++) nets[r] = new Ctrnn3LayersHebb(  );
    }
    else if ( !strcmp ( type_of_controller, "simple_three_layers" ) ){
        for(int r=0; r < num_agents; r++) nets[r] = new Simple3Layers(  );
    }
    else if ( !strcmp ( type_of_controller, "perceptron" ) ){
        for(int r=0; r < num_agents; r++) nets[r] = new Perceptron(  );
    }
    else if ( !strcmp ( type_of_controller, "elman" ) ){
        for(int r=0; r < num_agents; r++) nets[r] = new Elman(  );
    }
    else if ( !strcmp ( type_of_controller, "ctrnn_fully_recurrent" ) ){
        for(int r=0; r < num_agents; r++) nets[r] = new CtrnnFullyRecurrent (  );
    }
    else{
        cerr << " in Parameters::init_controllers - Controller type not found " << endl;
        exit(0);
    }
}

/* ---------------------------------------- */

void Parameters::init_ga( const int num_allele_per_solution, const int num_bases_per_allele ){

    if ( !strcmp ( type_of_ga, "roulette_wheel" ) ){ printf("create ga...\n");
        ga = new Roulette_wheel ( num_allele_per_solution, num_bases_per_allele, num_agents );
        /*
     char file_dir[500];
     sprintf(file_dir, "/home/elt7/Simulations/task_switch/EXP/GENOME/");
     char file_name[500];
     sprintf(file_name, "H");
     ga->upload_genome_from_file( file_dir, file_name, 0 );
    */

    }
    else if ( !strcmp ( type_of_ga, "tournament" ) ){
        ga = new Tournament ( num_allele_per_solution, num_bases_per_allele, num_agents );
    }
    else if ( !strcmp ( type_of_ga, "swarm_ga" ) ){
        ga = new Swarm_ga ( num_allele_per_solution, num_bases_per_allele, num_agents );
    }
    else{
        cerr << " in Parameters::init_controllers - GA type not found " << endl;
        exit(0);
    }
}

/* ---------------------------------------- */

void Parameters::dump_simulation_seed( void ){
    char fileName[500];
    sprintf(fileName, "%s%s.seed", rootSeedFileDir, rootSeedFileName);
    ofstream outfile ( fileName, ios::app);
    outfile.setf(ios::fixed);
    outfile << "ROOT_SEED=" << root_seed << endl;
    outfile.close();
}

/* ---------------------------------------- */

void Parameters::reset_seed( void ){
    GSL_randon_generator::reset_seed( root_seed );
}

/* ---------------------------------------- */
